﻿Option Strict On : Option Explicit On : Option Infer On

Imports System.Collections.ObjectModel
Imports System.ComponentModel
Imports System.IO
Imports System.IO.Pipes
Imports System.Text
Imports Microsoft.VisualBasic.ApplicationServices



Public Class SingleInstanceListener


   Public Event StartUpNextInstance(ByVal sender As Object, ByVal ev As StartupEventArgs)


   Private m_ServerPipe As NamedPipeServerStream

   Private m_sync As AsyncOperation

   Private m_Name As String

   Private Delegate Sub ListenerDelegate()




   Public Sub New()
      Me.New(GetAppIDName(Reflection.Assembly.GetExecutingAssembly))
   End Sub



   Public Sub New(ByVal name As String)
      m_Name = name

      ' m_sync is used to allow the StartUpNextInstance event to be raised on the origianl context.
      m_sync = AsyncOperationManager.CreateOperation(Nothing)

      If Not IsFirstInstance() Then
         ' create a client and notify the server
         Notify(GetArgsAsXML)
         End ' exit the applciation
      End If

   End Sub




   ' checks to see if a named pipe already exists. 
   ' If not initializes it.
   '
   Private Function IsFirstInstance() As Boolean

      If m_ServerPipe Is Nothing Then
         If InitServer() Then
            ' start the listener on a new thread via Delegate.BeginInvoke
            Dim listener As ListenerDelegate = AddressOf Listen
            listener.BeginInvoke(Nothing, Nothing)
         End If
      End If

      Return m_ServerPipe IsNot Nothing

   End Function




   ' safely initialize the named pipe, disposing of any exisiting one
   '
   Private Function InitServer() As Boolean
      Try
         If m_ServerPipe IsNot Nothing Then m_ServerPipe.Dispose()
         m_ServerPipe = New NamedPipeServerStream(m_Name)
         Return True
      Catch ex As Exception
         m_ServerPipe = Nothing
      End Try
      Return False
   End Function




   ' pipe server listening loop
   '
   Private Sub Listen()
      Do
         m_ServerPipe.WaitForConnection() 'thread pauses till connected

         Using sr As New StreamReader(m_ServerPipe)
            Dim strIn As String
            Dim sb As New StringBuilder
            Do
               strIn = sr.ReadLine
               If strIn Is Nothing Then Exit Do
               sb.Append(strIn)
            Loop
            ' use AsynOperation.Post to invoke on original context
            m_sync.Post(New Threading.SendOrPostCallback(AddressOf OnStartUpNextInstanceCallback), GetArgsFromXML(sb.ToString))
         End Using

      Loop While InitServer() 're-initialize the pipe server

   End Sub




   ' client pipe stream sends message to the server pipe stream
   '
   Private Sub Notify(ByVal message As String)
      Using pipeStream As New NamedPipeClientStream(m_Name)
         pipeStream.Connect()
         Using sw As New StreamWriter(pipeStream)
            sw.AutoFlush = True
            sw.WriteLine(message)
         End Using
      End Using
   End Sub



   ' called from the Listen method via a AsyncOperation.Post
   ' and raises the event onthe original context to allow UI interaction.
   '
   Private Sub OnStartUpNextInstanceCallback(ByVal args As Object)
      Dim ev As New StartupEventArgs(DirectCast(args, ReadOnlyCollection(Of String)))
      RaiseEvent StartUpNextInstance(Me, ev)
   End Sub




#Region "XML helper functions"

   '
   '  parses the xml string and creates a ReadOnlyCollection(of String) 
   '  with the items being the commandline arguements.
   '
   Private Function GetArgsFromXML(ByVal xmlString As String) As ReadOnlyCollection(Of String)
      Dim el As XElement = XElement.Parse(xmlString)
      Return New ReadOnlyCollection(Of String)((From item In el.<item> Select item.Value).ToArray)
   End Function



   ' returns the command line arguments as a xml string
   ' in the form of:
   ' <items><item>arg1</item><item>arg2</item></items>
   '
   Private Function GetArgsAsXML() As String
      Dim innerItems = From item In Environment.GetCommandLineArgs _
                  Select <item><%= item %></item>
      Dim items = <items><%= innerItems.Skip(1) %></items>
      Return items.ToString
   End Function


#End Region



#Region "helper functions"

   ' gets a name for the pipe equivalent to the name used by the VB single instance framework
   '
   Private Shared Function GetAppIDName(ByVal assemb As Reflection.Assembly) As String
      Dim typeLibGuidForAssembly As Guid = Runtime.InteropServices.Marshal.GetTypeLibGuidForAssembly(assemb)
      Dim strArray As String() = assemb.GetName.Version.ToString.Split("."c)
      Return (typeLibGuidForAssembly.ToString & strArray(0) & "." & strArray(1))
   End Function

#End Region


End Class
